/*
	cs.cc - classes to implement a client/server application

	(c) 1998 Polarware

	This file is distributed under the terms of the
	"GNU Library General Public License" (see file COPYING.LIB)
*/

#include "cs.h"
#include <algo.h> // find()

/*

  CSResource

 */

CSResource::CSResource( const string& _name ) 
	: type( -1 ), name( _name ), locked( false ), locker( NULL )
{}

void
CSResource::Read( Socket& s )
{
    s >> name >> locked;
    ReadData( s );
}

void
CSResource::Write( Socket& s )
{
    s << type << name << locked;
    WriteData( s );
}


/*
  
  Server

*/

Server::Server ( Socket::Port port )
	: login( port )
{
    login.Listen( 10 );
}

Server::~Server()
{
    unsigned i;
    for( i = 0; i < clients.size(); ++i )
	delete clients[i];
    for( i = 0; i < res.size(); ++ i )
	delete res[i];
    for( i = 0; i < protos.size(); ++ i )
	delete protos[i];
}

void
Server::RegisterType( const CSResource* r )
{
    protos.push_back( r );
}

void
Server::ProcessRequests()
{
    vector<Socket*>::iterator i;
    bool poll;
    CSService service;

    do {
	poll = CheckLogin();
	
	for (i = clients.begin(); i < clients.end(); i++)
	    if ( ( *i )->DataAvail() )
	     {
		 poll = true;
		 **i >> service;
		 switch( service )
		  {
		      case 'SHUT': Disconnect( *i ); break;
		      case 'SEND': Send( *i ); break;
		      case 'SALL': SendAll( *i ); break;
		      case 'NUMB': Number( *i ); break;
		      case 'NMEX': NameExists( *i ); break;
		      case 'LOCK': Lock( *i ); break;
		      case 'UNLK': Unlock( *i ); break;
		      case 'DELE': Delete( *i ); break;
		      default:	   InvalidRequest( *i, service ); break;	   
		  }
	     }
    } while( poll );
}

void
Server::Disconnect( Socket* s )
{
    clients.erase( find( clients.begin(), clients.end(), s ) );
}

void Server::Send( Socket* s )
{
    string n = s->ReadString();
    
    CSResource* r = (CSResource*) FindName( n );
    CSStatus status;
    
    if( r ) {
	status = csOK;
	s->Write( &status, sizeof( status ) );
	r->SetLocked( r->Locker() == s );
	r->Write( *s );
     }
     else {
	 status = csNOTFOUND;
	 s->Write( &status, sizeof( status ) );
     }
}

void
Server::SendAll( Socket* s )
{
    *s << (unsigned long) res.size();

    vector<CSResource*>::iterator i;
    for( i = res.begin(); i < res.end(); ++i ) {
	( *i )->SetLocked( (*i)->Locker() == s );
	( *i )->Write( *s );
    }
}

void
Server::Number( Socket* s )
{
    *s << (unsigned long) res.size();
}

void
Server::NameExists( Socket* s )
{
    string n = s->ReadString();
    char res = FindName( n ) != NULL;
    *s << res;
}

void
Server::Lock( Socket* s )
{
    string n = s->ReadString();
    CSStatus status;
    CSResource* r = (CSResource*) FindName( n );
    
    if( r )
	status = (!r->Locker() || r->Locker() == s) ? csOK : csLOCKERROR;
     else
	 status = csNOTFOUND;
    
    if( status == csOK )
	r->SetLocker( s );
    
    s->Write( &status, sizeof( status ) );
}

void
Server::Unlock( Socket* s )
{
    CSResource* r = ReadType( *s )->Clone( "" );
    r->Read( *s );
    r->SetLocked( false );
    
    CSStatus status = csOK;
    
    CSResource* found = (CSResource*) FindName( r->GetName() );
    if( found ) {
	    //* check lock **********************
	if( !found->Locker() || s == found->Locker() )
	    ReplaceRes( found, r );
	 else {
	     delete r;
	     status = csLOCKERROR;
	 }
    } else
	    //* name doesn't exist, add ************
	res.push_back( r );

    s->Write( &status, sizeof( status ) );
}

void Server::Delete( Socket* s )
{
    string n( s->ReadString() );    
    CSStatus status;

    CSResource* r = (CSResource*) FindName( n );
    if( r ) {
	 if( !r->Locker() || r->Locker () == s ) {
	     res.erase( find( res.begin(), res.end(), r ) );
	     status = csOK;
	 }
	  else
	      status = csLOCKERROR;
    } else
	status = csNOTFOUND;
    
    s->Write( &status, sizeof( status ) );
}

void
Server::InvalidRequest( Socket* s, CSService )
{
    Disconnect( s );
	// merciless, but it is difficult to recover from protocol errors
}

const CSResource*
Server::FindName( const string& n ) const
{
    vector<CSResource*>::const_iterator i;
    for( i = res.begin(); i < res.end(); ++i )
	if( (*i)->GetName() == n )
	    return *i;

    return NULL;
}    

const CSResource*
Server::ReadType( Socket& s ) const
{
    CSType type;
    s >> type;
    
    vector<const CSResource*>::const_iterator i;
    for( i = protos.begin(); i < protos.end(); ++i )
	if( (*i)->GetType() == type )
	    return *i;
    
    throw ExcCSNoPrototype( "Server::ReadType( Socket& )" );
}

void
Server::ReplaceRes( CSResource* from, CSResource* to )
{
    vector<CSResource*>::iterator i;
    for( i = res.begin(); i < res.end(); ++i )
	if( *i == from ) {
	    res.erase( i );
	    res.insert( i, to );
	}
}

bool
Server::CheckLogin()
{
    Socket* s;
    bool gotLogin = false;
    
    login.SetBlocking( false );
    if( ( s = login.Accept() ) != NULL ) {
	 clients.push_back( s );
	 gotLogin = true;
    }
    login.SetBlocking( true );
    return gotLogin;
}

/*

  Client

*/

Client::Client( const string& url, Socket::Port port ) 
	: server( port )
{
    try {
	if( !server.Connect( url.c_str() ) )
	    throw ExcCSCannotConnect( "Client::Client( const string&, Socket::Port )" );
    }
    catch( ExcSocketHostName ) {
	throw ExcCSCannotConnect( "Client::Client( const string&, Socket::Port )" );
    }   
}

Client::~Client ()
{
    if( server.Connected() ) {
	SaveAll ();
	 
	CSService service = 'SHUT';
	server << service;
	server.Close();
    }
}

void
Client::RegisterType( const CSResource* r )
{
    protos.push_back( r );
}

CSStatus Client::Add( CSResource* r )
{
    if( FindName( r->GetName() ) )
	return csNAMEEXISTS;

	//* check if name already exists on server *
    CSService service = 'NMEX';
    char nameExists;
    
    server << service << r->GetName();
    server >> nameExists; 
    if( nameExists )
	return csNAMEEXISTS;
    
    r->SetLocked( true );
    res.push_back( r );
    Unlock( r->GetName() );
    return csOK;
}

CSStatus
Client::Del( const string& s )
{
	// search resource *************************
    CSResource* r = NULL;
    vector<CSResource*>::iterator i;
    for( i = res.begin(); i < res.end(); ++i )
	if( (*i)->GetName() == s ) {
	    r = *i;
	    break;
	}
    
    if( r ) {
	    // request deletion from server ***
	CSService service = 'DELE';
	CSStatus status;
	server << service << r->GetName();
	server.Read( &status, sizeof( status ) );
	
	if( status == csOK )
	    res.erase( i );

	if( status == csNOTFOUND )
	    throw ExcCSResourceNotFound( "Client::Del( const string& )" );

	return status;
    }
     else
	 return csNOTFOUND;
}

CSStatus
Client::Load( const string& resname )
{
    CSService service = 'SEND';
    CSStatus status;
    CSResource* r;
    CSType type;

    server << service << resname;
    server.Read( &status, sizeof( status ) );
    if( status == csOK ) {
	server >> type;
	r = FindName( resname );
	if( !r )
	    Add( r = FindType( type )->Clone( "" ) );
	 
	r->Read( server );
    }
    
    return status;
}

CSStatus
Client::Lock( const string& resname )
{
    CSService service = 'LOCK';
    CSStatus status = Load( resname );
    
    if( status != csOK )
	return status;
    
    CSResource* r = FindName( resname );
    
    server << service << resname;
    
    server.Read( &status, sizeof( status ) );
    r->SetLocked( status == csOK );
    return status;
}

CSStatus Client::Unlock( const string& resname )
{
    CSResource* r = FindName( resname );
    CSStatus status;
    CSService service = 'UNLK';

    if( r ) {
	if( !r->IsLocked() )
	    return csLOCKERROR;
	 
	server << service;
	
	r->Write( server );
	server.Read( &status, sizeof( status ) );
	if( status == csLOCKERROR )
	    throw ExcCSUnlockFailed( "Client::Unlock()" );
	if( status == csNOTFOUND )
	    throw ExcCSResourceNotFound( "Client::Unlock()" );
	
	r->SetLocked( false );
	
	return status;
    }
     else
	 return csNOTFOUND;
}

void
Client::LoadAll()
{
    unsigned long num, i;
    CSType type;
    CSResource* r;
    CSService service = 'SALL';

    res.erase( res.begin(), res.end () );

    server << service;
    server >> num;
    
    for( i = 0; i < num; ++i ) {
	server >> type;
	r = FindType( type )->Clone( "" );
	res.push_back( r );
	r->Read( server );
    }
}

void
Client::SaveAll()
{
    vector<CSResource*>::iterator i;
    for( i = res.begin(); i < res.end(); ++i )
	if( (*i)->IsLocked() )
	    Unlock( (*i)->GetName() );
}

CSResource*
Client::FindName( const string& n ) const
{
    vector<CSResource*>::const_iterator i;
    for( i = res.begin(); i < res.end(); ++i )
	if( (*i)->GetName() == n )
	    return *i;
    return NULL;
}    
    
const CSResource*
Client::FindType( CSType type ) const
{
    vector<const CSResource*>::const_iterator i;
    for( i = protos.begin(); i < protos.end(); ++i )
	if( (*i)->GetType() == type )
	    return *i;
    
    throw ExcCSNoPrototype( "Client::ReadType( Socket& )" );
}
