/*
  cs.h - Classes to implement a client/server system

  (c) 1998 Martin Pitt (Piware)

  This file is distributed under the terms and conditions of the 
  "GNU Library General Public License" (see file COPYING.LIB)

  Please report bugs and comments to piware@piware.de
*/

#ifndef CS_H
#define CS_H

#include "socket_class.h"
#include <vector>

class istream;
class ostream;

typedef unsigned long CSService;

    /**
     * Since you cannot send objects through a network, an object
     * needs an identification of its type. Every derived CSResource
     * class must have an UNIQUE number that is stored in a CSType var.
     */
typedef long int CSType;

    /**
     * Return value of all failable client methods. csOK: all OK;
     * csNOTFOUND: resource was not found locally (respectively on
     * server if returned by Client::Load() or Lock() );
     * csNAMEEXISTS: (Client::Add() only) Name already exists;
     * csLOCKERROR: Resource is already locked by someone else
     */
enum CSStatus { csOK,  csNOTFOUND, csNAMEEXISTS, csLOCKERROR };

class Server;
class Client;

    /**
     * Abstract base class for all resources (objects) that can be
     * managed by CS. Derived classes have to provide a constructor,
     * Clone(), ReadData(), and WriteData().
     */
class CSResource
{
 public:
	/**
	 * Initializes the resource and sets its name. Derived classes
	 * must set the protected variable 'CSType type', which must
	 * be unique (per class, not per object) within the whole
	 * application. THIS IS EXTREMELY IMPORTANT!
	 * @param name Name of the resource; must be unique if you
	 *        want to insert it to a client object (which is
	 *        probably what you want)
	 */
    CSResource( const string& name );

	/// Destructor. Does nothing.
    virtual ~CSResource() {}
    
	/**
	 * A complete object cannot be sent through a socket, only its
	 * data and CSType. To make the receiver able to reconstruct
	 * the object, it must clone a prototype of the same CSType.
	 * All derived classes must provide this method! It must
	 * allocate a clone of itself, but with the name newName.
	 * (A precise copy is actually not neccessary to use the CS
	 * system, but since it exists, it could be used for other
	 * purposes, too.)
	 */
    virtual CSResource* Clone( const string& newName ) const = 0;

	/// Returns the resource's name.
    string GetName() const;
    
	/// Returns the resource's CSType.
    CSType GetType() const;

	/**
	 * Returns whether or not the resource is locked.
	 * @return true: This resource is locked to the caller's
	 *         client, so it may be modified; all other users
	 *         (clients) are not able to modify this resource.
	 *         false: This resource is not locked to you (either
	 *         not locked at all, or to someone else).
	 */
    bool IsLocked() const;

 protected:
	/// The class-unique CSType of this resource.
    CSType type;

	/**
	 * This method must write all specific resource data
	 * (i. e. all data but the name and CSType) to the given
	 * socket. It must correspond to the ReadData() method.
	 * @see ReadData()
	 */
    virtual void WriteData( Socket& ) = 0;

	/**
	 * This method must read all specific resource data
	 * (i. e. all data but the name and CSType) from the given
	 * socket. It must correspond to the WriteData() method.
	 * @see WriteData()
	 */
    virtual void ReadData( Socket& ) = 0;

 private:
    friend Server;
    friend Client;

    void Read (Socket&); // does not read type!
    void Write (Socket&); // does write type!

    void SetLocked (bool l) { locked = l; }
    void SetLocker (Socket* l) { locker = l; }
    Socket* Locker () const { return locker; }

 protected:
    string name;
    
 private:
    bool locked; // for clients only (r/o; writing does only confuse client,
	         // you cannot cheat the server!)
    Socket* locker; // for servers only 
};

    /**
     * This class is the manager of all resources, their locking state
     * and lockers. All classes that are to be used with the CS system
     * have to be registered with RegisterType()!
     */
class Server 
{
 public:
	/**
	 * Initializes the resource list and creates a socket
	 * listening at the specified port (where clients can login at).
	 */
    Server( Socket::Port );

	/**
	 * Deletes the login socket, all resources and the
	 * prototypes.
	 */
    virtual ~Server();

	/**
	 * Adds a new resource prototype. This registration should be
	 * done immediately after the constructor call for all
	 * resource types the CS system can handle. This is neccessary
	 * for object reconstruction after its data has been read from
	 * a socket.
	 * @param proto An instance of the resource class; Its data
	 *        and content do not matter, only its CSType.
	 * @see CSResource::Clone()
	 */
    void RegisterType( const CSResource* proto );

	/**
	 * Processes all pending client requests and logins and
	 * returns if there's nothing to be done any more. This method
	 * must be called regularly and frequently in the server
	 * application.
	 * @exception ExcCSNoPrototype (thrown by the unlock service
	 *            subfunction)
	 */
    virtual void ProcessRequests();

	/// Returns the number of clients that are currently logged in. 
    unsigned NumClients();

 protected:

//* requests (service / parameters / answer) *******************************
// RD: TypeID, name, local lock state, data of Read/WriteData()
    
    void Disconnect (Socket*);	// 'SHUT' / none / none
    void Send (Socket*);	// 'SEND' / name / CSStatus (OK: RD follows)
    void SendAll (Socket*);	// 'SALL' / none / long num, RD1, RD2, ..., RDnum
    void Number (Socket*);	// 'NUMB' / none / long num
    void NameExists (Socket*);	// 'NMEX' / name / char (1: does exist, 0: does not)
    void Lock (Socket*);	// 'LOCK' / name / CSStatus (OK -> RD follows)
    void Unlock (Socket*);	// 'UNLK' /  RD  / CSStatus
    void Delete (Socket*);	// 'DELE' / name / CSStatus
    virtual void InvalidRequest (Socket*, CSService);

	//* helper functions ************************
    bool CheckLogin ();
    const CSResource* FindName (const string&) const;
    const CSResource* ReadType (Socket&) const;
    void ReplaceRes (CSResource*, CSResource*);
    
 protected:
    vector <CSResource*> res;
    vector <const CSResource*> protos;
    vector <Socket*> clients;
    Socket login;
};

    /**
     * CS system Client. Establishes and conducts communication with a
     * CSServer. All classes that are to be used with the CS system
     * have to be registered with RegisterType()! To have a resource
     * be managed by the CS system, it has to be added to the
     * client. To edit it later, you first have to lock it to
     * you. After modification, it should be unlocked again to allow
     * other clients to modify it.
     */
class Client
{
 public:
	/**
	 * Connects to the server at given host and port. After the
	 * constructor call all prototypes should be registered.
	 * @param hostname Name of the host the Server is at.
	 * @param port Port at host the Server listens at.
	 * @exception ExcCSCannotConnect Could not connect to
	 * specified server at specified port.
	 */
    Client( const string& hostname, Socket::Port port );

	/// Unlocks all locked resources and disconnects from server.
    virtual ~Client();
    
	/**
	 * Adds a new resource class prototype. This method should be
	 * called immediately after the constructor call for all
	 * resources the CS system shall handle.
	 */
    void RegisterType( const CSResource* );

	/**
	 * Deletes all local resources and reloads all available
	 * resources from the server (including locking states).
	 * @exception ExcCSNoPrototype The server sent a resource of a
	 *            class not registered at the client.
	 */
    void LoadAll();

	/**
	 * Unlocks all locked resources, thereby updating the server
	 * of all changes that were made.
	 * @exception ExcCSUnlockFailed
	 * @exception ExcCSResourceNotFound
	 */
    void SaveAll();

	/**
	 * Adds a new (locally created) resource and sends it to the
	 * server. After this call, the resource is UNLOCKED!
	 * @return csOK: all OK; csNAMEEXISTS: There is already a
	 *         resource with this name.
	 * @exception ExcCSUnlockFailed
	 */
    CSStatus Add( CSResource* );
    
	/**
	 * Deletes the resource with given name both locally and at
	 * the server. This is possible only if the resource is locked
	 * to you or not locked to anybody.
	 * @return csOK: all OK; csLOCKERROR: res could not be
	 *         deleted, because it is locked by someone else;
	 *         csNOTFOUND: there is not resource with given name.
	 * @exception ExcCSResourceNotFound
	 */
    CSStatus Del( const string& );

	/**
	 * Reloads the resource with given name from the server
	 * (including locking state). If it does not exist locally, it
	 * will be created and added.
	 * @return csOK: all OK; csNOTFOUND: there is no resource with
	 *         specified name.
	 * @exception ExcCSNoPrototype
	 */
    CSStatus Load( const string& );

	/**
	 * Reloads the resource with given name from the server
	 * and requests locking it. If it does not exist locally, it
	 * will be created and added.
	 * @return csOK: all OK; csNOTFOUND: there is no resource with
	 *         specified name; csLOCKERROR: resource could not be
	 *         locked (but was loaded anyway).
	 * @exception ExcCSNoPrototype
	 */
    CSStatus Lock( const string& );

	/**
	 * Unlocks the specified resource and updates all changes to
	 * the server.
	 * @return csOK: all OK; csNOTFOUND: there is no resource with
	 *         given name; csLOCKERROR: resource is not locked to you.
	 * @exception ExcCSUnlockFailed, ExcCSResourceNotFound
	 */
    CSStatus Unlock( const string& );
    
	/// Returns the resource with given name (NULL if it does not exist).
    CSResource* FindName( const string& ) const;
 
	/// Returns the number of resources the client currently manages.
    unsigned ResCount() const;
    
	/// Returns the resource with given index (NULL if index is invalid).
    CSResource* Res( unsigned ) const;
    
	/// Array operator to return the resource with given index (NULL if index is invalid).
    CSResource* operator[] ( unsigned ) const;

	/// Returns the resource with given name (NULL if it does not exist).
    CSResource* operator[] ( const string& ) const;

 protected:
    const CSResource* FindType (CSType type) const;
    
 private:
    Socket server;
    vector<CSResource*> res;
    vector <const CSResource*> protos;
};

/*

  Exceptions

*/

    /**
     * Thrown by Client constructor if the client is unable to connect
     * to the server -> check hostname and port, and that the server
     * is running!
     */
class ExcCSCannotConnect : public Exception
{
 public:
    ExcCSCannotConnect (const string& _where) 
	    : Exception (_where, "cannot connect to specified server",
			 Exception::NONFATAL) {}
};

    /**
     * Thrown by the server and client services if the server or
     * client sent an object with a CSType for which no prototype
     * exists, so the object can not be reconstructed -> All resources
     * have to be registered with RegisterType()!. 
     */
class ExcCSNoPrototype : public Exception
{
 public:
    ExcCSNoPrototype (const string& _where) :
	    Exception (_where, "received a Resource type "
		       "which no prototype object exists for",
		       Exception::IPE) {}
};

    /**
     * Thrown by the client services if a resource was marked as
     * locked locally, but not on the server. -> Something messed up
     * the 'locked' variable of the resource; this variable is only an
     * indication for the client, the server does NOT use it!
     */
class ExcCSUnlockFailed : public Exception
{
 public:
    ExcCSUnlockFailed (const string& _where) :
	    Exception (_where, "unlocking failed -> somebody messed"
		       " with client resource's lock bit?!",
		       Exception::IPE) {}
};

    /**
     * A resource exists locally on the client, but not on the
     * server. This indicates an internal program error (client
     * resource list was messed up) and can not normally happen,
     * because all added resources are immediately sent to the server.
     */
class ExcCSResourceNotFound : public Exception
{
 public:
    ExcCSResourceNotFound (const string& _where) :
	    Exception (_where, "Resource not found on server although"
		       " it exists on client", Exception::IPE) {}
};


/*

  Inline methods

*/

inline
string
CSResource::GetName() const
{ 
    return name;
}

inline
CSType
CSResource::GetType() const
{ 
    return type;
}

inline
bool
CSResource::IsLocked() const
{ 
    return locked;
}

inline
unsigned
Server::NumClients () 
{
    return clients.size ();
}

inline
unsigned
Client::ResCount() const
{ 
    return res.size();
}

inline    
CSResource*
Client::Res( unsigned i ) const
{
    return ( i < res.size() ) ? res[i] : 0;
}

inline    
CSResource*
Client::operator[] ( unsigned i ) const
{ 
    return Res( i );
}

inline
CSResource*
Client::operator[] ( const string& s ) const
{
    return FindName( s );
} 

#endif // !defined( CS_H )
