/*
  socket_class.h - Interface for network programming

  (c) 1998 Martin Pitt (Piware)

  This file is distributed under the terms and conditions of the 
  "GNU Library General Public License" (see file COPYING.LIB)

  Please report bugs and comments to piware@piware.de
*/

#ifndef SOCKET_CLASS_H
#define SOCKET_CLASS_H

#include <string.h>	 // strerror()
#include <string>
#include <errno.h>	 // errno
#include "except.h"

/**
 * Two Socket objects connected together form a TCP communication
 * channel between a "server" and a "client". These may be to programs
 * running on either two different or the same computer, or even two
 * parts of the same program. So a Socket is a more general
 * bidirectional pipe.
 */
class Socket 
{
 public:
	/// Represents a port.
    typedef int Port;
    
 public:
	/**
	 * Initializes the socket to the given port and sets it to
	 * "closed" state.
	 */
    Socket( Port );

	/**
	 * Closes the socket if it is still open (i. e. listening or
	 * connected) and deletes it.
	 */
    virtual ~Socket();

	/**
	 * Closes the socket if it is still open. This method is
	 *called automatically in the destructor.
	 */
    void Close();

	/**
	 * Returns the socket's descriptor for use with the standard
	 * UNIX file functions (read, write, etc.). This method should
	 * not be needed in normal use, since SOcket provides stream
	 * I/O for all standard datatypes. It is only provided for
	 * completeness.
	 */
    int FileDescriptor() const { return socketDesc; }

	/**
	 * Sets the blocking behavior.
	 * @param blocking true: Read operations wait until data is
	 * available and Accept() waits until a client connects. false:
	 * Read operations do not wait, but read to EOF and return
	 * immediately; Accept() returns NULL if no client currently
	 * wants to connect.
	 * @see IsBlocking()
	 */
    void SetBlocking( bool blocking ) const;

	/**
	 * Returns the blocking behavior.
	 * @see SetBlocking()
	 */
    bool IsBlocking() const;

	// server functions ******************
    
	/**
	 * Puts the socket into "listen" mode, i. e. enables it to
	 * accept client connections. So, this method declares the
	 * socket to act as a server. The socket listens at the port
	 * specified in the constructor.
	 * @param maxClients The maximum number of queued login
	 *        requests (logins are queued until they are accepted
	 *        with Accept() ).
	 * @exception exSocketSys
	 * @see Accept()
	 */
    void Listen (unsigned maxClients = 1);

	/**
	 * Accepts the next queued login request from a client and
	 * establishes the communication channel. In blocking mode,
	 * this method waits (potentially forever) until a client logs
	 * in. To use this method, the socket must be in "listen" mode.	 
	 * @return A new socket that is connected to the client. The
	 * original (the accepting) socket remains listening to accept
	 * further connections; NULL if there are no connection
	 * request and the socket is in non-blocking mode.
	 */
    Socket* Accept () const;

	// client functions ******************

	/**
	 * Connects the socket to a server (i. e. listening) socket
	 * residing at the given hostname and at the port specified in the
	 * constructor. This method must be used at the client side.
	 * @return true: Connection established; false: Connection
	 *         could not be established (wrong hostname, no listening
	 *         socket at specified port, connection refused by server, etc.)
	 * @exception ExcSocketHostName The given hostname could not
	 *            be resolved.
	 */
    bool Connect( const char* hostname );

	/**
	 * Same as Connect( const char*), but host is given by IP
	 * number instead of the name.
	 */
    bool Connect ( unsigned long ip );

	/**
	 * Same as Connect( unsigned long ), but host is given as a
	 * dotted quad IP (for chars) instead of a single unsigned long.
	 */
    bool Connect( unsigned char dq1, unsigned char dq2, 
		  unsigned char dq3, unsigned char dq4 );

	/// Returns whether the socket is listening.
    bool Listening() const { return state == LISTENING; }

	/// Returns whether the socket is connected.
    bool Connected() const { return state == CONNECTED; }

	/// Returns whether the socket is closed.
    bool Closed() const { return state == CLOSED; }

	/**
	 * Returns whether data is available for reading. This method
	 * is especially useful if socket is used in non-blocking
	 * mode. To use this method, the socket must be connected!
	 */
    bool DataAvail() const;

	/**
	 * Waits up to the specified timeout for incoming data.
	 * @param timeOut Maximum number of seconds to wait for data.
	 * @return true: There is data to read from the socket; false:
	 * there is still no data available.
	 */
    bool WaitRead( unsigned timeOut ) const;

	/**
	 * Converts the given IP number to a dotted quad format
	 * suitable to print out.
	 */
    static string DottedQuadIP( unsigned long );

	/// Determines the IP number of given hostname.
    static unsigned long Resolve( const char* );

	/// Determines the hostname of given IP number.
    static string Resolve( unsigned long );

	/**
	 * Reads data from a connected socket.
	 * @param buf Buffer for read data (must be at least len bytes big).
	 * @param len Number of bytes to read.
	 * @return Actual number of read bytes; equal to len if in
	 * blocking mode, but can be smaller if socket does not block.
	 */
    unsigned Read( void* buf, unsigned len ) const;

	/// Reads a string.
    string ReadString() const;

	/// Reads an unsigned long, caring for network to host byte order conversion.
    unsigned long  ReadLong() const;

	/// Reads an unsigned short, caring for network to host byte order conversion.
    unsigned short ReadShort() const;

	/// Reads a float, caring for network to host byte order conversion.
    float ReadFloat() const;

	/// Reads a double, caring for network to host byte order conversion.
    double ReadDouble() const;

	/// Reads a char.
    char ReadChar() const;
    
	/// Reads a bool.
    bool ReadBool() const;

	/**
	 * Writes data to a connected socket.
	 * @param buf The data to send. Must be at least len bytes big.
	 * @param len Number of bytes to write.
	 */
    void Write( const void* buf, unsigned len ) const;

	/// Writes a string.
    void WriteString( const string& ) const;

	/// Writes an unsigned long, caring for host to network byte order conversion.
    void WriteLong( unsigned long ) const;

	/// Writes an unsigned short, caring for host to network byte order conversion.
    void WriteShort( unsigned short ) const;

	/// Writes a float, caring for host to network byte order conversion.
    void WriteFloat( float ) const;

	/// Writes a double, caring for host to network byte order conversion.
    void WriteDouble( double ) const;

	/// Writes a char.
    void WriteChar( char ) const;

	/// Writes a bool.
    void WriteBool( bool ) const;

	/// Stream operator wrapper for ReadString().
    const Socket& operator >> (string&) const;
	/// Stream operator wrapper for ReadLong().
    const Socket& operator >> (unsigned long&) const;
	/// Stream operator wrapper for ReadLong(), but signed.
    const Socket& operator >> (long&) const;
	/// Stream operator wrapper for ReadShort().
    const Socket& operator >> (unsigned short&) const;
	/// Stream operator wrapper for ReadShort(), but signed.
    const Socket& operator >> (short&) const;
	/// Stream operator wrapper for ReadLong(), but signed.
    const Socket& operator >> (int&) const;
	/// Stream operator wrapper for ReadFloat().
    const Socket& operator >> (float&) const;
	/// Stream operator wrapper for ReadDouble().
    const Socket& operator >> (double&) const;
	/// Stream operator wrapper for ReadChar().
    const Socket& operator >> (char&) const;
	/// Stream operator wrapper for ReadBool().
    const Socket& operator >> (bool&) const;

	/// Stream operator wrapper for WriteString().
    const Socket& operator << (const string&) const;
	/// Stream operator wrapper for WriteLong().
    const Socket& operator << (const unsigned long) const;
	/// Stream operator wrapper for WriteLong(), but signed.
    const Socket& operator << (const long) const;
	/// Stream operator wrapper for WriteShort().
    const Socket& operator << (const unsigned short) const;
	/// Stream operator wrapper for WriteShort(), but signed.
    const Socket& operator << (const short) const;
	/// Stream operator wrapper for WriteLong().
    const Socket& operator << (const int) const;
	/// Stream operator wrapper for WriteFloat().
    const Socket& operator << (const float) const;
	/// Stream operator wrapper for WriteDouble().
    const Socket& operator << (const double) const;
	/// Stream operator wrapper for WriteChar().
    const Socket& operator << (const char) const;
	/// Stream operator wrapper for WriteBool().
    const Socket& operator << (const bool) const;
    
 private:
    Socket (int _socketDesc, Port _port);
    
    int socketDesc;
    Port port;
    
    enum SocketState {CLOSED, LISTENING, CONNECTED} state;
};

//* Exceptions *****************************************************

class ExcSocketSys : public Exception
{
 public:
    ExcSocketSys( const char* _where ) :
	    Exception( _where, strerror( errno ) ) {}
};

class ExcSocketHostName : public Exception
{
 public:
    ExcSocketHostName( const char* _where ) :
	    Exception( _where, "Invalid server host name", Exception::NONFATAL ) {}
};

//* INLINE METHODS **************************************************

inline
bool
Socket::Connect( const char* hostname )
{
    return Connect( Resolve( hostname ) );
}

inline
bool
Socket::ReadBool() const
{
    return bool (ReadChar() );
}

inline
void
Socket::WriteBool( bool b ) const
{
    WriteChar( b ? 1 : 0 ); // assure that bool value fits into one char
}

inline
const Socket&
Socket::operator >> ( unsigned long &x ) const
{
    x = ReadLong(); return *this;
}

inline
const Socket&
Socket::operator >> (long &x) const
{
    x = long( ReadLong () ); return *this;
}

inline
const Socket&
Socket::operator >> ( unsigned short &x ) const
{
    x = ReadShort(); return *this;
}

inline
const Socket&
Socket::operator >> ( short &x ) const
{
    x = short( ReadShort () ); return *this;
}

inline
const Socket&
Socket::operator >> ( int &x ) const
{
    x = int( ReadLong () ); return *this;
}

inline
const Socket&
Socket::operator >> ( float &x ) const
{
    x = ReadFloat(); return *this;
}

inline
const Socket&
Socket::operator >> ( double &x ) const
{
    x = ReadDouble(); return *this;
}

inline
const Socket&
Socket::operator >> ( char &x ) const
{
    x = ReadChar(); return *this;
}

inline
const Socket&
Socket::operator >> ( bool &x ) const
{
    x = ReadBool(); return *this;
}

inline
const Socket&
Socket::operator << ( const string& x ) const
{
    WriteString( x ); return *this;
}

inline
const Socket&
Socket::operator << ( const unsigned long x ) const
{
    WriteLong( x ); return *this;
}

inline
const Socket&
Socket::operator << ( const long x ) const
{
    WriteLong( (unsigned long) x ); return *this;
}

inline
const Socket&
Socket::operator << ( const unsigned short x ) const
{
    WriteShort( x ); return *this;
}

inline
const Socket&
Socket::operator << ( const short x ) const
{
    WriteShort( x ); return *this;
}

inline
const Socket&
Socket::operator << ( const int x ) const
{
    WriteLong( x ); return *this;
}

inline
const Socket&
Socket::operator << ( const float x ) const
{
    WriteFloat( x ); return *this;
}

inline
const Socket&
Socket::operator << ( const double x ) const
{
    WriteDouble( x ); return *this;
}

inline
const Socket&
Socket::operator << ( const char x ) const
{
    WriteChar( x ); return *this;
}

inline
const Socket&
Socket::operator << ( const bool x ) const
{
    WriteBool( x ); return *this;
}

#endif /* !defined ( SOCKET_CLASS_H ) */



